from remotemanager.serialisation.serialyaml import serialyaml
from remotemanager.serialisation.serialjson import serialjson
import remotemanager.serialisation.serial  # noqa: F401, this silences an IDE warning

__all__ = ["serialyaml", "serialjson"]

try:
    from remotemanager.serialisation.serialdill import serialdill  # noqa: F401

    __all__.append("serialdill")
except ImportError:
    pass

try:
    from remotemanager.serialisation.serialjsonpickle import (  # noqa: F401
        serialjsonpickle,
    )

    __all__.append("serialjsonpickle")
except ImportError:
    pass



from remotemanager import SendableMixin, LoggingMixin


class placeholder_option(SendableMixin, LoggingMixin):
    """
    .. warning::
        This class is intended to be subclassed by the optional and required
        placeholders.

    Stub class to sit in place of an option within a computer.

    Args:
        mode (string):
            storage mode, required or optional
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
    """

    def __init__(self, mode, flag, min, max):
        self._mode = mode
        self._flag = flag
        self._value = None
        self._bool = False

        self._min = min
        self._max = max

    def __hash__(self):
        return hash(self._mode)

    def __str__(self):
        return str(self.value)

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None

    @property
    def flag(self):
        return self._flag

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def value(self):
        if hasattr(self, "default") and self._value is None:
            return self.default

        return self._value

    @value.setter
    def value(self, value):
        try:
            value / 1
            isnumeric = True
        except TypeError:
            isnumeric = False

        if isnumeric:
            if self.min is not None and value < self.min:
                raise ValueError(
                    f"{value} for {self.flag} " f"is less than minimum value {self.min}"
                )
            if self.max is not None and value > self.max:
                raise ValueError(
                    f"{value} for {self.flag} " f"is more than maximum value {self.max}"
                )

        self._bool = True
        self._value = value


class required(placeholder_option):
    """
    This option is _required_, and should raise an error if no value is found

    Args:
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
    """

    def __init__(self, flag, min=None, max=None):
        super().__init__("required", flag, min, max)


class optional(placeholder_option):
    """
    This option is not required, and should have an accessible default if
    no value is found

    Args:
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
        default:
            default value to use if none is assigned
    """

    def __init__(self, flag, default=None, min=None, max=None):
        super().__init__("optional", flag, min, max)

        self._default = default

    @property
    def default(self):
        try:
            return self._default()
        except TypeError as E:
            self._logger.warning(f"recieved error from default attempt: {E}")
            return self._default



"""
Handles file transfer via `cp`
"""

from remotemanager.transport.transport import Transport


class cp(Transport):
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._logger.info("created new cp transport")

        self._cmd = "mkdir -p {secondary} ; " "cp -r --preserve {primary} {secondary}"

    def cmd(self, primary, secondary):
        base = self._cmd.format(primary=primary, secondary=secondary)
        self._logger.debug(f'returning formatted cmd: "{base}"')
        return base



"""
URL base class for connecting to remote systems
"""
import logging
import os
import time
import typing
from collections import deque

from remotemanager.connection.testing_object import ConnectionTest
from remotemanager.logging.utils import format_iterable
from remotemanager.connection.cmd import CMD
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import ensure_list
from remotemanager.logging import LoggingMixin
from remotemanager.logging.verbosity import Verbosity


class URL(SendableMixin, LoggingMixin):
    """
    Container to store the url info for a Remote run

    The url should contain everything pertaining to the _remote_, allowing
    Dataset to be remote-agnostic

    Arguments:
        user (str):
            username for the remote system
        host (str):
            host address of the remote system
        port (int, str):
            port to connect to for ssh tunnels
        verbose (bool):
            base-verbosity for connections
        timeout (int):
            time to wait before issuing a timeout for cmd calls
        max_timeouts (int):
            number of times to attempt cmd communication in case of a timeout
        python (str):
            string used to initiate a python instance
        raise_errors (bool):
            set false to ignore errors by default in cmd calls
        passfile (str):
            absolute path to password file for sshpass calls
        envpass (str):
            environment variable containing absolute path to password file for
            sshpass calls
        cmd_history_depth (int):
            number of cmd calls to store history for
        kwargs:
            any extra args that may end up here from a Dataset or Computer are
            discarded
    """

    _localhost = "localhost"

    _do_not_package = ["_urlutils"]

    _defaults = {"python": "python", "submitter": "bash", "shell": "bash"}

    def __init__(
        self,
        user: str = None,
        host: str = None,
        port: int = None,
        verbose: [int, Verbosity] = None,
        timeout: int = 5,
        max_timeouts: int = 3,
        python: str = None,
        submitter: str = None,
        shell: str = None,
        raise_errors: bool = True,
        keyfile: str = None,
        passfile: str = None,
        envpass: str = None,
        cmd_history_depth: int = 10,
        **kwargs,
    ):

        self._verbose = Verbosity(verbose)

        if host is None:
            host = URL._localhost
        self._conn = {"user": user, "host": host, "port": port}

        self.timeout = timeout
        self.max_timeouts = max_timeouts
        self._python = python
        self._submitter = submitter
        self._shell = shell

        self._cmd_history_depth = cmd_history_depth
        self._cmd_history = deque(maxlen=self._cmd_history_depth)

        # explicit path takes precedent over environment variable
        if passfile is None and envpass is not None:
            passfile = os.environ[envpass]

        if "ignore_errors" in kwargs:
            raise_errors = not kwargs.pop("ignore_errors")

        self._keyfile = keyfile
        self._passfile = passfile
        self._raise_errors = raise_errors
        self._ssh_override = None
        self.ssh_prepend = None

        self._logger.info(
            f"new url created with url details:{format_iterable(self._conn)}"
        )

        self._callcount = 0

        self._latency = 0
        self._connection_test = None

    @property
    def raise_errors(self):
        return self._raise_errors

    @raise_errors.setter
    def raise_errors(self, r):
        self._raise_errors = r

    @property
    def verbose(self) -> Verbosity:
        """
        Return the current verbosity setting

        Returns:
            (Verbosity): current verbosity
        """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose) -> None:
        """
        Verbosity setter
        """
        self._verbose = Verbosity(verbose)

    @property
    def call_count(self) -> int:
        return self._callcount

    @property
    def user(self) -> str:
        """
        Currently configured username
        """
        return self._conn["user"]

    @user.setter
    def user(self, user):
        """
        Set the user attribute
        """
        self._conn["user"] = user

    @property
    def host(self) -> str:
        """
        Currently configured hostname
        """
        return self._conn["host"] or URL._localhost

    @host.setter
    def host(self, host):
        """
        Set the host attribute
        """
        self._conn["host"] = host

    @property
    def userhost(self) -> str:
        """
        `user@host` string if possible, just `host` if user is not present
        """
        if self.user is None:
            return self.host
        else:
            return f"{self.user}@{self.host}"

    @property
    def python(self):
        if self._python is not None:
            return self._python
        return URL._defaults["python"]

    @python.setter
    def python(self, python):
        self._python = python

    @property
    def submitter(self):
        if self._submitter is not None:
            return self._submitter
        if self._shell is not None:
            self._logger.debug("submitter is None but shell is set, returning shell")
            return self._shell
        return URL._defaults["submitter"]

    @submitter.setter
    def submitter(self, submitter):
        self._submitter = submitter

    @property
    def shell(self):
        if self._shell is not None:
            return self._shell
        return URL._defaults["shell"]

    @shell.setter
    def shell(self, shell):
        self._shell = shell

    @property
    def port(self) -> int:
        """
        Currently configured port (defaults to 22)
        """
        port = self._conn["port"] or 22
        return port

    @property
    def ssh(self) -> str:
        """
        ssh insert for commands on this connection
        """
        if self.is_local:
            return ""

        if self._ssh_override:
            return self._ssh_override

        ret = []

        if self.passfile is not None:
            ret.append(f"sshpass -f {self.passfile}")

        ret.append(f"ssh -p {self.port}")

        if self._keyfile:
            ret.append(f"-i {self.keyfile}")

        ret.append(self.userhost)

        return " ".join(ret)

    @ssh.setter
    def ssh(self, newssh: str) -> None:
        """
        Allows forced override of the ssh command

        Inserting extra flags into the ssh can be done as follows:

        >>> url = URL()
        >>> print(url.ssh)
        >>> "ssh"
        >>> url.ssh = "LANG=C " + url.ssh
        >>> print(url.ssh)
        >>> "LANG=C ssh"

        Args:
            newssh (str):
                new ssh string to insert

        Returns:
            None
        """
        self._ssh_override = newssh

    def clear_ssh_override(self):
        """
        Wipe any override applied to ssh. Can also be done by setting
        url.ssh = None

        Returns:
            None
        """
        self._ssh_override = None

    @property
    def keyfile(self):
        if self._keyfile is None:
            return
        p = self._keyfile.replace("~", os.environ["HOME"])
        if not os.path.isfile(p):
            raise RuntimeError(f"could not find ssh key file at {self._keyfile}")

        return self._keyfile

    @property
    def passfile(self):
        if self._passfile is None:
            return

        p = self._passfile.replace("~", os.environ["HOME"])
        if not os.path.isfile(p):
            raise RuntimeError(f"could not find password file at {self._passfile}")

        return self._passfile

    @property
    def is_tunnel(self):
        """
        Uses the presence of an assigned port to determine if we are tunnelling
        or not

        Returns (bool):
           True if this is a tunnelled connection
        """
        return self._conn["port"] is not None

    @property
    def is_local(self):
        """
        True if this connection is purely local
        """
        host = self.host
        if host == URL._localhost:
            return True
        elif host.startswith("127."):
            return True
        return False

    def ping(self, n: int = 5, timeout: int = 30) -> float:
        """
        Perform and monitor a ping command

        Args:
            n (int): number of pings to aim for
            timeout (int): kill the process and return 0 if
                this period is elapsed

        Returns:
            (float) latency in ms
        """
        self._logger.runtime(f"pinging {self.host}")

        def cleanup(process):
            process.kill()
            try:
                os.remove("tmp_ping")
                self._logger.debug("removed temporary file")
            except FileNotFoundError:
                self._logger.debug("temporary ping file could not be found")

        ping = self.cmd(
            f"ping {self.host}", local=True, stdout="tmp_ping", asynchronous=True
        )

        lines = []
        t0 = time.time()
        while len(lines) < n + 1:
            with open("tmp_ping", "r") as o:
                lines = o.readlines()

            time.sleep(0.5)

            if time.time() - t0 > timeout:
                self._logger.warning(f"ping timed out after {timeout} seconds")
                cleanup(ping)
                return -1

        cleanup(ping)

        times = []
        for line in lines[1:]:
            times.append(self._process_ping_line(line))

        avg = sum(times) / n
        self._logger.runtime(f"ping times: {times} -> {avg}")
        return avg

    def _process_ping_line(self, line: str) -> float:
        """
        Function to parse the contents of a line output from a ping command

        Args:
            line (str):
                ping command line
        Returns:
            (float): ping time in seconds
        """
        timing = line.split("time=")[1].strip()  # get timing
        val, units = timing.split()
        self._logger.debug(f"parsed line {line} -> {val} ({units})")

        if units != "ms":
            raise ValueError(f"Unknown ping units {units}")

        return float(val) * 1e-3  # convert to seconds

    def cmd(
        self,
        cmd: str,
        asynchronous: bool = False,
        local: [bool, None] = None,
        stdout: str = None,
        stderr: str = None,
        timeout: int = None,
        max_timeouts: int = None,
        raise_errors: bool = None,
        dry_run: bool = False,
        prepend: bool = False,
        force_file: bool = False,
    ) -> typing.Union[CMD, str]:
        """
        Creates and executes a command

        Args:
            asynchronous (bool):
                run this command asynchronously
            cmd (str):
                command to execute
            local (bool, None):
                force a local or remote execution. Defaults to None
            stdout (str):
                optional file to redirect stdout to
            stderr (str):
                optional file to redirect stderr to
            timeout (int):
                time to wait before issuing a timeout
            max_timeouts (int):
                number of times to attempt communication in case of a timeout
            raise_errors (bool):
                override for global setting. Raise any stderr if encountered
            dry_run (bool):
                don't exec the command if True, just returns the string
            prepend (bool):
                always attempt to use ssh_prepend
            force_file (bool):
                passthrough for CMD force_file argument

        Returns (CMD):
            returned command instance
        """
        _remote_call = False
        if local is not None and not local:
            self._logger.info("forced remote call; appending ssh to cmd")
            cmd = f"{self.ssh} '{cmd}'"
            _remote_call = True
        elif local is None and not self.is_local:
            self._logger.info("implicit remote call; appending ssh to cmd")
            cmd = f"{self.ssh} '{cmd}'"
            _remote_call = True
        if raise_errors is None:
            raise_errors = self._raise_errors

        timeout = self.timeout if timeout is None else timeout
        max_timeouts = self.max_timeouts if max_timeouts is None else max_timeouts

        if self.ssh_prepend is not None:
            self._logger.info(f"found prepend string {self.ssh_prepend}")
            prep = self.ssh_prepend.strip()

            if prepend:
                self._logger.debug(f"force adding prepend string {prep}")
                cmd = f"{prep} {cmd}"
            elif _remote_call:
                self._logger.debug(f"implicit adding prepend string {prep}")
                cmd = f"{prep} {cmd}"

        if dry_run:
            return cmd.strip()

        thiscmd = CMD(
            cmd.strip(),
            asynchronous=asynchronous,
            stdout=stdout,
            stderr=stderr,
            timeout=timeout,
            max_timeouts=max_timeouts,
            raise_errors=raise_errors,
            force_file=force_file,
        )
        thiscmd.exec()
        if not local:
            self._callcount += 1

        self._cmd_history.append(thiscmd)

        return thiscmd

    @property
    def cmd_history(self):
        return self._cmd_history

    @property
    def cmd_history_depth(self):
        return self._cmd_history_depth

    @cmd_history_depth.setter
    def cmd_history_depth(self, newdepth: int):
        """
        Updates the history depth, and creates a new dequeue populated with as many of
        the existing cmds as possible

        Args:
            newdepth (int):
                new depth to capture
        Returns:

        """
        self._logger.info(
            f"updating history depth from {self.cmd_history_depth} to {newdepth}"
        )
        self._cmd_history_depth = newdepth

        newqueue = deque(maxlen=self.cmd_history_depth)

        for item in self.cmd_history:
            newqueue.append(item)

        self._cmd_history = newqueue

    @property
    def utils(self):
        """
        Handle for the URLUtils module
        """
        if self.is_missing("_urlutils"):
            self._urlutils = URLUtils(self)
        return self._urlutils

    def test_connection(self) -> None:
        """
        Create a ConnectionTest instance and run the tests

        Returns:
            None
        """
        testing_object = ConnectionTest(self)
        testing_object.exec()

        self._connection_test = testing_object

    @property
    def connection_test(self) -> ConnectionTest:
        """
        Return the connection test object

        Returns:
            ConnectionTest: testing object
        """
        return self._connection_test

    @property
    def connection_data(self) -> dict:
        """
        Returns the results of a previous connection test

        Returns:
            (dict) connection data
        """
        return self._connection_test.data

    @property
    def latency(self) -> [float, None]:
        """
        Attempts to access the latency property of the stored ConnectionTest

        Returns:
            (float): connection latency in seconds, if available. Else None
        """
        if self._connection_test is not None:
            return self._connection_test.latency
        return None

    def script(self, **kwargs):
        raise NotImplementedError


class URLUtils:
    """
    Extra functions to go with the URL class, called via URL.utils

    As it requires a parent `URL` to function, and is instantiated with a
    `URL`, there is little to no purpose to using this class exclusively

    Arguments:
        parent (URL):
            parent class to provide utils to
    """

    def __init__(self, parent: URL):

        self._logger = logging.getLogger(__name__ + ".URLUtils")
        self._logger.info(f"creating a utils extension to parent: {parent}")

        self._parent = parent

    def file_mtime(
        self,
        files: list,
        local: bool = None,
        python: bool = False,
        ignore_empty: bool = False,
        dry_run: bool = False,
    ) -> dict:
        """
        Check file modification times of [files]

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            python (bool):
                ensure python style search is used
            ignore_empty (bool):
                also check the filesize, ignoring empty files
            dry_run (bool):
                print command only

        Returns (dict):
            {file: mtime (unix)} dictionary
        """

        self._logger.info(f"performing stat on files: {files}")
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)
        times, error = self._file_mtime(files, local, python, dry_run)

        if dry_run:
            # in this instance "times" is simply the command
            return times

        self._logger.info("received:")
        self._logger.info(times)
        self._logger.info(error)
        output = {}
        for file in files:
            if file in times:

                mtime = times[file][0]
                fsize = times[file][1]
                if ignore_empty and fsize == 0:
                    output[file] = None
                else:
                    output[file] = mtime
            else:
                output[file] = None

        return output

    def _file_mtime(self, files: list, local: bool, python: bool, dry_run: bool):
        """
        Perform the "stat -c %Y" command on a list of files,
        returning the result. Uses a python command backup if this fails

        Args:
            files (list):
                list of files to check
            local (bool):
                force a local search
            python (bool):
                force the python override
            dry_run (bool):
                print command only

        Returns:
            (list): list of file unix times
        """
        sep = ","

        def stat():
            self._logger.debug("attempting raw stat command on files")
            basecmd = f"stat -c %n{sep}%Y{sep}%s"
            if len(files) == 1:
                cmd = f"{basecmd} {files[0]}"
            else:
                cmd = f"{basecmd} {{" + ",".join(files) + "}"

            ret = self._parent.cmd(
                cmd, local=local, raise_errors=False, dry_run=dry_run
            )

            if dry_run:
                return ret, "", "", ""

            times = {}
            for line in ret.stdout.split("\n"):
                try:
                    fname = line.split(sep)[0]
                    mtime = int(float(line.split(sep)[1]))
                    fsize = int(float(line.split(sep)[2]))

                    times[fname] = (mtime, fsize)
                except IndexError:
                    pass

            return times, ret.stderr.split("\n"), ret.returncode, ret.stderr

        def pystat():
            self._logger.debug("attempting python stat on files")
            ex = f"""import os
files={files}
for f in files:
\ttry: print(f'{{f}}{sep}{{os.stat(f).st_mtime}}{sep}{{os.stat(f).st_size}}')
\texcept FileNotFoundError: print(f)"""

            cmd = f'{self._parent.python} -c "{ex}"'

            ret = self._parent.cmd(
                cmd, local=local, raise_errors=False, dry_run=dry_run
            )

            if dry_run:
                return ret, None

            times = {}
            error = []
            for line in ret.stdout.split("\n"):
                try:
                    fname = line.split(sep)[0]
                    mtime = int(float(line.split(sep)[1]))
                    fsize = int(float(line.split(sep)[2]))

                    times[fname] = (mtime, fsize)
                except IndexError:
                    error.append(line.strip())

            return times, error

        files = ensure_list(files)

        if not python:
            t, e, returncode, stderr = stat()
            if returncode in [126, 127] or "illegal option" in stderr:
                self._logger.warning("stat failed, falling back on python")
                return pystat()

            return t, e

        return pystat()

    def file_presence(
        self, files: list, local: bool = None, dry_run: bool = False
    ) -> dict:
        """
        Search for a list of files, returning a boolean presence dict

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        self._logger.info(f"checking for presence of files: {files}")
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)

        times = self.file_mtime(files, local=local, dry_run=dry_run)

        if dry_run:
            return times

        return {f: times[f] is not None for f in files}

    def search_folder(
        self, files: list, folder: str, local: bool = None, dry_run: bool = False
    ) -> dict:
        """
        Search `folder` for `files`, returning a boolean presence dict

        Arguments:
            files (list):
                list of filenames to check for. Optionally, a string for a
                single file
            folder (str):
                folder to scan
            local (bool):
                perform the scan locally (or remotely)
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        if local is None:
            local = self._parent.is_local
        fpath = os.path.abspath(folder) if local else folder

        self._logger.debug(f"scanning folder {fpath}")
        self._logger.debug("searching for files:")
        self._logger.debug(f"{format_iterable(files)}")

        ls_return = self.ls(fpath, local=local, as_list=True, dry_run=dry_run)

        if dry_run:
            self._logger.info("dry run, returning command")
            return ls_return

        scan = [os.path.basename(f) for f in ls_return]

        self._logger.debug("scan sees:")
        self._logger.debug(f"{format_iterable(scan)}")

        if isinstance(files, str):
            self._logger.info("files is a string, running in singular mode")
            ret = {files: os.path.basename(files) in scan}
        else:
            ret = {file: os.path.basename(file) in scan for file in files}

        return ret

    def touch(
        self,
        file: str,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> CMD:
        """
        perform unix `touch`, creating or updating `file`

        Arguments:
            file (str):
                filename or path to file
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f"utils touch on file {file}")
        fname = os.path.abspath(file) if local else file
        return self._parent.cmd(
            f"touch {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

    def mkdir(
        self,
        file: str,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> CMD:
        """
        perform unix `mkdir -p`, creating a folder structure

        Arguments:
            file (str):
                name or path to folder
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f"utils mkdir on path {file}")
        fname = os.path.abspath(file) if local else file
        # print(f'making dir {fname}')
        return self._parent.cmd(
            f"mkdir -p {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

    def ls(
        self,
        file: str,
        as_list: bool = True,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> [CMD, list]:
        """
        Identify the files present on the directory

        Arguments:
            file (str):
                name or path to folder.
            as_list (bool):
                convert to a list format
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD, list):
            CMD instance for the command, or the list if as_list is True
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f"utils ls on path {file}")
        fname = os.path.abspath(file) if local else file

        ret = self._parent.cmd(
            f"ls {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

        if as_list and not dry_run:
            ret = [f for f in ret.stdout.split("\n") if f != ""]
        return ret



import time

from remotemanager.logging import LoggingMixin
from remotemanager.storage import SendableMixin

import os

from remotemanager.utils import dir_delta


class TrackedFile(LoggingMixin, SendableMixin):

    __slots__ = ("_remote_path", "_local_path", "_file")

    def __init__(self, local_path, remote_path, file):

        self._remote_path = remote_path
        self._local_path = local_path
        self._file = file

        self._last_seen = {"remote": -1, "local": -1}

    def __repr__(self):
        return self.local

    def __fspath__(self):
        return self.name

    @property
    def name(self):
        return self._file

    @property
    def importstr(self):
        return os.path.splitext(self._file)[0]

    @property
    def remote(self):
        return os.path.join(self._remote_path, self.name)

    @property
    def local(self):
        return os.path.join(self._local_path, self.name)

    @property
    def remote_dir(self):
        return self._remote_path

    @property
    def local_dir(self):
        return self._local_path

    def relative_remote_path(self, other: str) -> str:
        """
        Return a path relative to `cwd`

        Args:
            other:
                working dir to compare against

        Returns:
            relative path
        """
        # if our remote path is an abspath, we already have what we need
        if os.path.isabs(self.remote_dir):
            return self.remote

        # we're already in the remote, just return the filename
        if self.remote_dir == other:
            return self.name

        # find the deepest shared path, treat it as a "root"
        stem = os.path.commonpath([self.remote_dir, other])
        # find how far down this stem is from `other`
        dirdelta = dir_delta(stem, other)
        # generate a ../ string that steps "down" to the common path
        down = "../" * dirdelta

        tmp_remote = self.remote_dir.replace(stem, "").strip("/")
        # rebuild up from our virtual root
        return os.path.join(down, tmp_remote, self.name)

    @property
    def content(self):
        with open(self.local, "r") as o:
            self.confirm_local()
            return o.read()

    def write(self, content: str) -> None:
        """
        Write `content` to the local copy of the file

        Args:
            content:
                content to write
        Returns:
            None
        """
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        with open(self.local, "w+") as o:
            o.write(content)
        self.confirm_local()

    def append(self, content: str) -> None:
        """
        Append `content` to the local copy of the file

        Args:
            content:
                content to append
        Returns:
            None
        """
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        with open(self.local, "a+") as o:
            o.write(content)
        self.confirm_local()

    def confirm_local(self):
        """
        Confirm sighting of the file locally
        """
        self._last_seen["local"] = int(time.time())

    def confirm_remote(self):
        """
        Confirm sighting of the file on the remote
        """
        self._last_seen["remote"] = int(time.time())

    def last_seen(self, where: str) -> int:
        return self._last_seen[where]

    @property
    def last_seen_local(self):
        return self.last_seen("local")

    @property
    def last_seen_remote(self):
        return self.last_seen("remote")



import os
import warnings

from remotemanager.storage.remotefunction import cached_functions
from remotemanager.logging import LoggingMixin
from remotemanager.storage.sendablemixin import SendableMixin


class Dependency(SendableMixin, LoggingMixin):

    _do_not_package = ["_network"]

    def __init__(self):
        self._logger.info("new Dependency created")

        self._network = []
        self._parents = []
        self._children = []

    def add_edge(self, primary, secondary):
        pair = (primary, secondary)
        if pair not in self._network:
            self._logger.info(f"adding new edge {pair}")

            self._parents.append(primary.short_uuid)
            self._children.append(secondary.short_uuid)

            self._network.append(pair)

    @property
    def network(self):
        return self._network

    def get_children(self, dataset):
        uuid = dataset.short_uuid

        tmp = []
        for i in range(len(self._parents)):
            if self._parents[i] == uuid:
                tmp.append(self.network[i][1])

        return tmp

    def get_parents(self, dataset):
        uuid = dataset.short_uuid

        tmp = []
        for i in range(len(self._children)):
            if self._children[i] == uuid:
                tmp.append(self.network[i][0])

        return tmp

    @property
    def ds_list(self):

        datasets = []
        for pair in self.network:
            for ds in pair:
                if ds not in datasets:
                    datasets.append(ds)

        return datasets

    def remove_run(self, id: bool = False) -> bool:
        out = []
        for ds in self.ds_list:
            out.append(ds.remove_run(id=id, dependency_call=True))

        return all(out)

    def clear_runs(self) -> None:
        for ds in self.ds_list:
            ds.clear_runs(dependency_call=True)

    def clear_results(self, wipe) -> None:
        for ds in self.ds_list:
            ds.clear_results(wipe, dependency_call=True)

    def wipe_local(self, files_only: bool = False) -> None:
        for ds in self.ds_list:
            ds.wipe_local(files_only=files_only, dependency_call=True)

    def wipe_remote(self, files_only: bool = False) -> None:
        for ds in self.ds_list:
            ds.wipe_remote(files_only=files_only, dependency_call=True)

    def hard_reset(self, files_only: bool = False) -> None:
        for ds in self.ds_list:
            ds.hard_reset(files_only=files_only, dependency_call=True)

    def append_run(
        self, caller, chain_run_args, run_args, force, lazy, *args, **kwargs
    ):
        """
        Appends runs with the same args to all datasets

        Args:
            caller:
                (Dataset): The dataset which defers to the dependency
            chain_run_args (bool):
                for dependency runs, will not propagate run_args to other datasets in
                the chain if False (defaults True)
            run_args (dict):
                runner arguments
            *args:
                append_run args
            **kwargs:
                append_run keyword args

        Returns:
            None
        """
        self._logger.info(f"appending run from {caller}")

        datasets = self.ds_list
        self._logger.info(f"There are {len(datasets)} datasets in the chain")

        if chain_run_args:
            self._logger.info("chain_args is True, propagating")
            kwargs.update(run_args)

        for ds in datasets:
            if ds == caller:
                caller_args = {k: v for k, v in kwargs.items()}
                caller_args.update(run_args)
                ds.append_run(
                    dependency_call=True, force=force, lazy=lazy, *args, **caller_args
                )
            else:
                ds.append_run(
                    dependency_call=True, force=force, lazy=lazy, *args, **kwargs
                )

        for ds in datasets:
            parents = self.get_parents(ds)
            if len(parents) > 1:
                warnings.warn(
                    "Multiple parents detected. "
                    "Variable passing in this instance is unstable!"
                )
            for parent in parents:
                # TODO this is broken with multiple parents
                lstr = (
                    f"import os.path\n"
                    f'if os.path.getmtime("'
                    f'{parent.runners[-1].runfile.name}") > '
                    f'os.path.getmtime("'
                    f'{parent.runners[-1].resultfile.name}"):\n'
                    f'\traise RuntimeError("outdated '
                    f'result file for parent")\n'
                    f'repo.loaded = repo.load("'
                    f'{parent.runners[-1].resultfile.name}")'
                )
                ds.runners[-1]._dependency_info["parent_import"] = lstr

            tmp = []
            for child in self.get_children(ds):
                runner = child.runners[-1]
                tmp.append(
                    f"{child.url.submitter} {runner.jobscript.name} "
                    f"2>> {runner.errorfile.name}"
                )

            ds.runners[-1]._dependency_info["child_submit"] = tmp

            if not lazy:
                ds.database.update(ds.pack())

    def finish_append(self) -> None:
        for ds in self.ds_list:
            ds.finish_append(dependency_call=True)

    def run(
        self,
        dry_run: bool = False,
        **run_args,
    ):
        self._logger.info("dependency internal run call")

        ds_store = {}
        for ds in self.ds_list:
            ds_store[ds] = len(ds.runners)

        if not len(set(ds_store.values())) == 1:
            msg = f"Datasets do not have matching numbers of runners!: " f"{ds_store}"
            self._logger.critical(msg)
            raise RuntimeError(msg)

        # we need to write a common repo containing all functions
        first = list(ds_store.keys())[0]

        if not os.path.isdir(first.master_script.local_dir):
            os.makedirs(first.master_script.local_dir)

        with open(first.repofile.local, "w+") as o:
            if len(cached_functions) > 0:
                o.write("### cached functions ###\n")

                for fname, f in cached_functions.items():
                    o.write(f.source + "\n")
                o.write("\n")

            o.write("### primary functions ###\n")
            for ds in ds_store:
                o.write(ds.function.source)
                o.write("\n")

            o.write("\n\n### serialiser functions ###")
            o.write(first.serialiser.dumpfunc())
            o.write("\n")
            o.write(first.serialiser.loadfunc())

        first.transport.queue_for_push(first.repofile)

        # and now a master script to kick off the chains
        # we need the export for ALL datasets
        master_content = [f"export {first.main_dir_env}=$PWD\n"]

        for ds in ds_store:
            # reuse the store as a list of runner uuids for runners_to_update
            ds_store[ds] = []
            for runner in ds.runners:
                ready = runner.stage(
                    python=ds.url.python,
                    repo=f"${first.main_dir_env}/{first.repofile.name}",
                    **run_args,
                )
                if not ready:
                    continue

                # we only need the runlines for the first dataset, since the chaining
                # is handled by the jobscripts
                if ds == first:
                    submitter = ds.url.submitter
                    jobscript = runner.jobscript.relative_remote_path(
                        ds.master_script.remote_dir
                    )
                    jobpath, jobfile = os.path.split(jobscript)

                    runline = f"{submitter} {jobfile} " f"2>> {runner.errorfile.name}"

                    if ds.remote_dir != runner.remote_dir:
                        runline = f"cd {jobpath} && " + runline

                    asynchronous = runner.run_args["asynchronous"]
                    if asynchronous and submitter == "bash":
                        ds._logger.debug('appending "&" for async run')
                        runline += f" &"

                    master_content.append(runline)

                first.transport.queue_for_push(runner.jobscript)
                first.transport.queue_for_push(runner.runfile)

                for file in runner.extra_files["send"]:
                    first.transport.queue_for_push(
                        os.path.split(file)[1],
                        os.path.split(file)[0],
                        runner.remote_dir,
                    )

                ds_store[ds].append(runner.uuid)

        # finally write the script
        with open(first.master_script.local, "w+") as o:
            # need to remove any prior errors
            for dataset in ds_store:
                o.write(f"rm -f *{dataset.short_uuid}*error.out\n")
            o.write("\n")

            for line in master_content:
                o.write(line)
                if first.add_newline:
                    o.write("\n")

        first.transport.queue_for_push(first.master_script)

        if not dry_run:
            first.transport.transfer()

            cmd = (
                f"cd {first.remote_dir} && {first.url.shell} {first.master_script.name}"
            )
            first._run_cmd = first.url.cmd(cmd, asynchronous=True)

            for ds in ds_store:
                ds.set_runner_states("submit pending", ds_store[ds])
        else:
            first.transport.wipe_transfers()
            for ds in ds_store:
                ds.set_runner_states("dry run", ds_store[ds])
